/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.persistence.impl.util;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import kieker.analysis.analysisComponent.AbstractAnalysisComponent;
import kieker.analysis.model.analysisMetaModel.MIAnalysisComponent;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIFilter;
import kieker.analysis.model.analysisMetaModel.MIInputPort;
import kieker.analysis.model.analysisMetaModel.MIOutputPort;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProperty;
import kieker.analysis.model.analysisMetaModel.MIReader;
import kieker.analysis.model.analysisMetaModel.MIRepository;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.domain.pluginDecorators.AbstractAnalysisComponentDecorator;
import kieker.webgui.domain.pluginDecorators.FilterDecorator;
import kieker.webgui.domain.pluginDecorators.ReaderDecorator;
import kieker.webgui.domain.pluginDecorators.RepositoryDecorator;

import net.vidageek.mirror.dsl.Mirror;

import org.springframework.stereotype.Service;

/**
 * This service provides methods to convert a given {@link Class} instance (the one from a reflection loaded repository or plugin) to a meta model instance. This is
 * done by using the annotations and methods which will be accessed via the reflection API.
 * 
 * @author Nils Christian Ehmke
 */
@Service
public class Class2ModelInstanceConverter {

	private static final Log LOG = LogFactory.getLog(Class2ModelInstanceConverter.class);

	private static final MIAnalysisMetaModelFactory FACTORY = MIAnalysisMetaModelFactory.eINSTANCE;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public Class2ModelInstanceConverter() {
		// No code necessary
	}

	public ReaderDecorator convertReaderClass2ModelInstance(final Class<AbstractReaderPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		return (ReaderDecorator) this.convertComponentClass2ModelInstance(clazz, classAndMethodContainer, Type.Reader);
	}

	public FilterDecorator convertFilterClass2ModelInstance(final Class<AbstractFilterPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		return (FilterDecorator) this.convertComponentClass2ModelInstance(clazz, classAndMethodContainer, Type.Filter);
	}

	public RepositoryDecorator convertRepositoryClass2ModelInstance(final Class<AbstractRepository> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		return (RepositoryDecorator) this.convertComponentClass2ModelInstance(clazz, classAndMethodContainer, Type.Repository);
	}

	@SuppressWarnings("unchecked")
	private AbstractAnalysisComponentDecorator<? extends MIAnalysisComponent> convertComponentClass2ModelInstance(
			final Class<? extends AbstractAnalysisComponent> clazz,
			final ClassAndMethodContainer classAndMethodContainer, final Type type) {
		final MIAnalysisComponent plugin = this.createSuitableModelInstance(clazz, classAndMethodContainer);

		final Collection<MIProperty> properties = new ArrayList<MIProperty>();
		final Collection<MIInputPort> inputPorts = new ArrayList<MIInputPort>();
		final Collection<MIOutputPort> outputPorts = new ArrayList<MIOutputPort>();
		final Collection<MIDisplay> displays = new ArrayList<MIDisplay>();
		final Map<String, String> propertyDescriptions = new HashMap<String, String>();
		final Map<String, String> displayDescriptions = new HashMap<String, String>();
		String description = "";
		String dependency = "";
		boolean fullyInitialized = true;

		try {
			description = this.fillDescription(clazz, classAndMethodContainer);
			dependency = this.fillDependency(clazz, classAndMethodContainer);

			this.fillProperties(clazz, classAndMethodContainer, properties, propertyDescriptions);
			plugin.getProperties().addAll(properties);
			if ((type == Type.Filter) || (type == Type.Reader)) {
				this.fillOutputPorts((Class<AbstractPlugin>) clazz, classAndMethodContainer, outputPorts, (MIPlugin) plugin);
				this.fillDisplays((Class<AbstractPlugin>) clazz, classAndMethodContainer, displays, displayDescriptions);
				((MIPlugin) plugin).getOutputPorts().addAll(outputPorts);
				((MIPlugin) plugin).getDisplays().addAll(displays);
				if (type == Type.Filter) {
					this.fillInputPorts((Class<AbstractFilterPlugin>) clazz, classAndMethodContainer, inputPorts, (MIFilter) plugin);
					((MIFilter) plugin).getInputPorts().addAll(inputPorts);
				}
			}
		} catch (final Throwable ex) { // NOCS (Throwable)
			Class2ModelInstanceConverter.LOG.info("A component with the classname '" + clazz.getCanonicalName() + "' could not be initialized.", ex);

			plugin.getProperties().clear();
			if ((type == Type.Filter) || (type == Type.Reader)) {
				((MIPlugin) plugin).getOutputPorts().clear();
				((MIPlugin) plugin).getDisplays().clear();
				if (type == Type.Filter) {
					((MIFilter) plugin).getDisplays().clear();
				}
			}

			fullyInitialized = false;
		}
		plugin.setClassname(clazz.getName());
		plugin.setName(clazz.getSimpleName());

		switch (type) {
		case Filter: {
			return new FilterDecorator((MIFilter) plugin, Collections.unmodifiableMap(propertyDescriptions), description, dependency, fullyInitialized,
					Collections.unmodifiableMap(displayDescriptions));
		}
		case Reader: {
			return new ReaderDecorator((MIReader) plugin, Collections.unmodifiableMap(propertyDescriptions), description, dependency, fullyInitialized,
					Collections.unmodifiableMap(displayDescriptions));
		}
		case Repository: {
			return new RepositoryDecorator((MIRepository) plugin, Collections.unmodifiableMap(propertyDescriptions), description, dependency, fullyInitialized);

		}
		default:
			return null;
		}
	}

	private void fillInputPorts(final Class<AbstractFilterPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer,
			final Collection<MIInputPort> inputPorts,
			final MIFilter parent) {
		final Collection<Method> methods = this.getInputPortMethods(clazz, classAndMethodContainer);
		for (final Method method : methods) {
			final Annotation inputPortAnnotation = method.getAnnotation(classAndMethodContainer.getInputPortAnnotationClass());
			final MIInputPort newPort = Class2ModelInstanceConverter.FACTORY.createInputPort();
			newPort.setName((String) new Mirror().on(inputPortAnnotation).invoke().method("name").withoutArgs());
			newPort.setParent(parent);
			inputPorts.add(newPort);
		}
	}

	private void fillDisplays(final Class<? extends AbstractPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer,
			final Collection<MIDisplay> displays,
			final Map<String, String> displayDescriptions) {
		final Collection<Method> methods = this.getDisplayMethods(clazz, classAndMethodContainer);
		for (final Method method : methods) {
			final Annotation displayAnnotation = method.getAnnotation(classAndMethodContainer.getDisplayAnnotationClass());
			final MIDisplay newDisplay = Class2ModelInstanceConverter.FACTORY.createDisplay();
			newDisplay.setName((String) new Mirror().on(displayAnnotation).invoke().method("name").withoutArgs());
			displays.add(newDisplay);
			displayDescriptions.put(newDisplay.getName(), (String) new Mirror().on(displayAnnotation).invoke().method("description").withoutArgs());
		}
	}

	private Collection<Method> getAnnotatedMethods(final Class<? extends AbstractPlugin> clazz, final Class<? extends Annotation> annotation) {
		final Collection<Method> result = new ArrayList<Method>();

		final Method[] methods = clazz.getMethods();
		for (final Method method : methods) {
			if (method.isAnnotationPresent(annotation)) {
				result.add(method);
			}
		}

		return result;
	}

	private Collection<Method> getInputPortMethods(final Class<AbstractFilterPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		return this.getAnnotatedMethods(clazz, classAndMethodContainer.getInputPortAnnotationClass());
	}

	private Collection<Method> getDisplayMethods(final Class<? extends AbstractPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		return this.getAnnotatedMethods(clazz, classAndMethodContainer.getDisplayAnnotationClass());
	}

	@SuppressWarnings("unchecked")
	private void fillOutputPorts(final Class<? extends AbstractPlugin> clazz, final ClassAndMethodContainer classAndMethodContainer,
			final Collection<MIOutputPort> outputPorts, final MIPlugin parent) {
		final Annotation annotation = this.getSuitableAnnotation(clazz, classAndMethodContainer);
		final Annotation[] outputPortAnnotations = (Annotation[]) new Mirror().on(annotation).invoke().method("outputPorts").withoutArgs();
		for (final Annotation outputPortAnnotation : outputPortAnnotations) {
			final MIOutputPort newPort = Class2ModelInstanceConverter.FACTORY.createOutputPort();
			newPort.setName((String) new Mirror().on(outputPortAnnotation).invoke().method("name").withoutArgs());
			newPort.setParent(parent);
			outputPorts.add(newPort);
		}
	}

	@SuppressWarnings("unchecked")
	private void fillProperties(final Class<? extends AbstractAnalysisComponent> clazz, final ClassAndMethodContainer classAndMethodContainer,
			final Collection<MIProperty> properties, final Map<String, String> propertyDescriptions) {
		final Annotation annotation = this.getSuitableAnnotation(clazz, classAndMethodContainer);
		final Annotation[] propertyAnnotations = (Annotation[]) new Mirror().on(annotation).invoke().method("configuration").withoutArgs();
		for (final Annotation propertyAnnotation : propertyAnnotations) {
			final MIProperty newProperty = Class2ModelInstanceConverter.FACTORY.createProperty();
			newProperty.setName((String) new Mirror().on(propertyAnnotation).invoke().method("name").withoutArgs());
			newProperty.setValue((String) new Mirror().on(propertyAnnotation).invoke().method("defaultValue").withoutArgs());
			properties.add(newProperty);
			propertyDescriptions.put(newProperty.getName(), (String) new Mirror().on(propertyAnnotation).invoke().method("description").withoutArgs());
		}
	}

	private String fillDependency(final Class<? extends AbstractAnalysisComponent> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final Annotation annotation = this.getSuitableAnnotation(clazz, classAndMethodContainer);
		return (String) new Mirror().on(annotation).invoke().method("dependencies").withoutArgs();
	}

	private String fillDescription(final Class<? extends AbstractAnalysisComponent> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final Annotation annotation = this.getSuitableAnnotation(clazz, classAndMethodContainer);
		return (String) new Mirror().on(annotation).invoke().method("description").withoutArgs();
	}

	private Annotation getSuitableAnnotation(final Class<? extends AbstractAnalysisComponent> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		// Get the two potential annotations
		final Annotation annotationPlugin = clazz.getAnnotation(classAndMethodContainer.getPluginAnnotationClass());
		final Annotation annotationRepository = clazz.getAnnotation(classAndMethodContainer.getRepositoryAnnotationClass());

		// Find out which of them exists
		if (annotationPlugin != null) {
			return annotationPlugin;
		} else {
			return annotationRepository;
		}
	}

	private MIAnalysisComponent createSuitableModelInstance(final Class<? extends AbstractAnalysisComponent> clazz,
			final ClassAndMethodContainer classAndMethodContainer) {
		if (classAndMethodContainer.getAbstractReaderPluginClass().isAssignableFrom(clazz)) {
			return Class2ModelInstanceConverter.FACTORY.createReader();
		} else if (classAndMethodContainer.getAbstractFilterPluginClass().isAssignableFrom(clazz)) {
			return Class2ModelInstanceConverter.FACTORY.createFilter();
		} else {
			return Class2ModelInstanceConverter.FACTORY.createRepository();
		}
	}

	/**
	 * Tells whether the given class is "programmaticOnly". If there is no annotation which could tell the state of the flag, true will be returned.
	 * 
	 * @param clazz
	 *            The class of the plugin or repository.
	 * @param classAndMethodContainer
	 *            The container which will be used for the reflection access.
	 * 
	 * @return The state of the programmaticOnly flag of the plugin or repository.
	 */
	public boolean isProgrammaticOnly(final Class<? extends AbstractAnalysisComponent> clazz, final ClassAndMethodContainer classAndMethodContainer) {
		final Annotation annotation = this.getSuitableAnnotation(clazz, classAndMethodContainer);
		return (Boolean) new Mirror().on(annotation).invoke().method("programmaticOnly").withoutArgs();
	}

	private static enum Type {
		Reader, Filter, Repository
	}

}
